import 'dart:io';
import 'dart:mirrors';

import 'package:args/args.dart';

void main(List<String> arguments) {
  new TaskRunner().run(arguments);
}

class TaskRunner {
  final LibraryMirror _library;
  final List<Task> tasks;

  factory TaskRunner() {
    final library = currentMirrorSystem().isolate.rootLibrary;

    final methods = library.declarations.values
        .where((d) =>
            !d.isPrivate &&
            d.isTopLevel &&
            d is MethodMirror &&
            MirrorSystem.getName(d.simpleName) != 'main')
        .toList();

    final tasks = methods.map((m) => new Task(m)).toList();

    return new TaskRunner._(tasks, library);
  }

  TaskRunner._(this.tasks, this._library);

  ArgParser get parser {
    final parser = new ArgParser();

    tasks.forEach((t) => parser.addCommand(t.name, t.parser));
    parser.addFlag('help', abbr: 'h', negatable: false);

    return parser;
  }

  void run(List<String> arguments) {
    final argResults = parser.parse(arguments);

    if (argResults['help']) {
      print(parser.usage);
      exit(1);
    }

    final command = argResults.command;
    final keys = command.options.where((s) => command.wasParsed(s));
    final namedArguments = new Map<Symbol, String>.fromIterable(keys,
        key: (k) => MirrorSystem.getSymbol(k), value: (k) => command[k]);

    _library.invoke(MirrorSystem.getSymbol(command.name), [], namedArguments);
  }
}

class Task {
  final String name;
  final List<String> options;
  final List<String> flags;

  factory Task(MethodMirror method) {
    final name = MirrorSystem.getName(method.simpleName);

    if (!method.parameters
        .every((p) => [bool, String].contains(p.type.reflectedType))) {
      throw 'Parameters should be String or bool';
    }

    final options = method.parameters
        .where((p) => p.type.reflectedType == String)
        .map((m) => MirrorSystem.getName(m.simpleName))
        .toList();

    final flags = method.parameters
        .where((p) => p.type.reflectedType == bool)
        .map((m) => MirrorSystem.getName(m.simpleName))
        .toList();

    return new Task._(name, options, flags);
  }

  Task._(this.name, this.options, this.flags);

  ArgParser get parser {
    final parser = new ArgParser();
    options.forEach((o) => parser.addOption(o));
    flags.forEach((f) => parser.addFlag(f));
    return parser;
  }
}
